import unittest
from pyqint import Molecule, HF
import numpy as np

class TestHF(unittest.TestCase):

    def test_hartree_fock_h2o(self):
        """
        Test Hartree-Fock calculation on water using STO-3G basis set
        """
        mol = Molecule()
        mol.add_atom('O', 0.0, 0.0, 0.0)
        mol.add_atom('H', 0.7570, 0.5860, 0.0)
        mol.add_atom('H', -0.7570, 0.5860, 0.0)

        results = HF().rhf(mol, 'sto3g')

        # check that energy matches
        np.testing.assert_almost_equal(results['energy'], -73.21447132, 4)

        # verify that terms are being calculated
        np.testing.assert_almost_equal(results['density'], np.einsum('ik,jk,k->ij', results['orbc'], results['orbc'], [2,2,2,2,2,0,0]), decimal=5)
        np.testing.assert_almost_equal(results['ekin'] + results['enuc'] + results['erep'] + results['ex'] + results['enucrep'], results['energy'], decimal=5)

    def test_hartree_fock_ch4(self):
        """
        Test Hartree-Fock calculation on water using STO-3G basis set
        """
        mol = Molecule()
        dist = 1.78/2
        mol.add_atom('C', 0.0, 0.0, 0.0, unit='angstrom')
        mol.add_atom('H', dist, dist, dist, unit='angstrom')
        mol.add_atom('H', -dist, -dist, dist, unit='angstrom')
        mol.add_atom('H', -dist, dist, -dist, unit='angstrom')
        mol.add_atom('H', dist, -dist, -dist, unit='angstrom')

        results = HF().rhf(mol, 'sto3g')

        # check that orbital energies are correctly approximated
        ans = np.array([-11.0707,
                         -0.7392,
                         -0.3752,
                         -0.3752,
                         -0.3752,
                          0.2865,
                          0.4092,
                          0.4092,
                          0.4092])
        np.testing.assert_almost_equal(results['orbe'], ans, 4)

        en = -39.35007843284954
        np.testing.assert_almost_equal(results['energies'][-1], en, 4)
        
    def test_hartree_fock_ch4_symmetric(self):
        """
        Test Hartree-Fock calculation on CH4 using an STO-3g basis set and
        symmetric orthogonalization
        """
        mol = Molecule()
        dist = 1.78/2
        mol.add_atom('C', 0.0, 0.0, 0.0, unit='angstrom')
        mol.add_atom('H', dist, dist, dist, unit='angstrom')
        mol.add_atom('H', -dist, -dist, dist, unit='angstrom')
        mol.add_atom('H', -dist, dist, -dist, unit='angstrom')
        mol.add_atom('H', dist, -dist, -dist, unit='angstrom')

        results = HF().rhf(mol, 'sto3g', ortho='symmetric')

        # check that orbital energies are correctly approximated
        ans = np.array([-11.0707,
                         -0.7392,
                         -0.3752,
                         -0.3752,
                         -0.3752,
                          0.2865,
                          0.4092,
                          0.4092,
                          0.4092])
        np.testing.assert_almost_equal(results['orbe'], ans, 4)

        en = -39.35007843284954
        np.testing.assert_almost_equal(results['energies'][-1], en, 4)

    def test_hartree_fock_restart(self):
        """
        Test Hartree-Fock calculation on water using STO-3G basis set
        """
        mol = Molecule()
        dist = 1.78/2
        mol.add_atom('C', 0.0, 0.0, 0.0, unit='angstrom')
        mol.add_atom('H', dist, dist, dist, unit='angstrom')
        mol.add_atom('H', -dist, -dist, dist, unit='angstrom')
        mol.add_atom('H', -dist, dist, -dist, unit='angstrom')
        mol.add_atom('H', dist, -dist, -dist, unit='angstrom')
        results1 = HF().rhf(mol, 'sto3g')

        # check that orbital energies are correctly approximated
        ans = np.array([-11.0707,
                         -0.7392,
                         -0.3752,
                         -0.3752,
                         -0.3752,
                          0.2865,
                          0.4092,
                          0.4092,
                          0.4092])
        np.testing.assert_almost_equal(results1['orbe'], ans, 4)

        en = -39.35007843284954
        np.testing.assert_almost_equal(results1['energies'][-1], en, 4)

        # create new CH4 molecule with slight adjustment in geometry and
        # seed the calculation with the previous converged result
        mol = Molecule()
        dist = 1.78/2
        mol.add_atom('C', 0.0, 0.0, 0.1, unit='angstrom')
        mol.add_atom('H', dist, dist, dist, unit='angstrom')
        mol.add_atom('H', -dist, -dist, dist, unit='angstrom')
        mol.add_atom('H', -dist, dist, -dist, unit='angstrom')
        mol.add_atom('H', dist, -dist, -dist, unit='angstrom')

        # perform HF calculation with coefficient matrix from previous
        # calculation to speed up the convergence
        results2 = HF().rhf(mol, 'sto3g', orbc_init=results1['orbc'])

        # assess that the energy of the perturbed result is different
        # (and also higher)
        en = -39.34538546003782
        np.testing.assert_almost_equal(results2['energies'][-1], en, 3)

        # check that the convergence is quicker
        self.assertTrue(len(results1['energies']) > len(results2['energies']))

if __name__ == '__main__':
    unittest.main()



import unittest
from pyqint import PyQInt, gto, Molecule
import numpy as np

class TestKinetic(unittest.TestCase):

    def test_gto_kinetic(self):
        """
        Test kinetic integrals for primitive GTOs

        Tij = <gto_I | -1/2 nabla^{2} | gto_j>
        """

        # construct integrator object
        integrator = PyQInt()

        # test GTO
        gto1 = gto(0.154329, [0.0, 0.0, 0.0], 3.425251, 0, 0, 0)
        gto2 = gto(0.535328, [0.0, 0.0, 0.0], 0.623914, 0, 0, 0)
        gto3 = gto(0.444635, [0.0, 0.0, 0.0], 0.168855, 0, 0, 0)
        kinetic = integrator.kinetic_gto(gto1, gto1)
        result = 1.595603108406067
        np.testing.assert_almost_equal(kinetic, result, 4)

    def test_cgf_kinetic(self):
        """
        Test kinetic integrals for contracted Gaussians

        Tij = <cgf_i | -1/2 nabla^{2} | cgf_j>
        """

        # construct integrator object
        integrator = PyQInt()

        # build hydrogen molecule
        mol = Molecule("H2")
        mol.add_atom('H', 0.0, 0.0, 0.0)
        mol.add_atom('H', 0.0, 0.0, 1.4)
        cgfs, nuclei = mol.build_basis('sto3g')

        T = np.zeros((2,2))
        T[0,0] = integrator.kinetic(cgfs[0], cgfs[0])
        T[0,1] = T[1,0] = integrator.kinetic(cgfs[0], cgfs[1])
        T[1,1] = integrator.kinetic(cgfs[1], cgfs[1])

        T11 = 0.7600315809249878
        T12 = 0.2364544570446014
        np.testing.assert_almost_equal(T[0,0], T11, 4)
        np.testing.assert_almost_equal(T[1,1], T11, 4)
        np.testing.assert_almost_equal(T[0,1], T12, 4)

if __name__ == '__main__':
    unittest.main()



import unittest
from pyqint import Molecule, HF, COHP
import numpy as np

class TestCOHP(unittest.TestCase):

    def testCO(self):
        """
        Test construction of localized orbitals using Foster-Boys procedure
        for the CO molecule
        """
        d = 1.145414
        mol = Molecule()
        mol.add_atom('C', 0.0, 0.0, -d/2, unit='angstrom')
        mol.add_atom('O', 0.0, 0.0,  d/2, unit='angstrom')

        res = HF().rhf(mol, 'sto3g')
        cohp = COHP(res).run(res['orbc'], 0, 1)

        cohp_ref = np.array([
            0.0399,
            0.0104,
           -0.4365,
            0.2051,
           -0.2918,
           -0.2918,
            0.1098,
            0.5029,
            0.5029,
            6.4827
        ])

        # note that Foster-Boys optimization is somewhat random and thus
        # we use relatively loose testing criteria
        np.testing.assert_almost_equal(cohp,
                                       cohp_ref,
                                       decimal=4)

if __name__ == '__main__':
    unittest.main()



from pytessel import PyTessel
from . import PyQInt
import numpy as np

def build_isosurface(filename, mol, basis, coeff, isovalue, dim = 5.0):
    sz = 99
    integrator = PyQInt()
    grid = integrator.build_rectgrid3d(-dim, dim, sz)
    scalarfield = np.reshape(integrator.plot_wavefunction(grid, coeff, basis), (sz, sz, sz))
    unitcell = np.diag(np.ones(3) * 10.0)

    pytessel = PyTessel()
    vertices_p, normals_p, indices_p = pytessel.marching_cubes(scalarfield.flatten(), scalarfield.shape, unitcell.flatten(), isovalue)
    vertices_n, normals_n, indices_n = pytessel.marching_cubes(scalarfield.flatten(), scalarfield.shape, unitcell.flatten(), -isovalue)

    build_abo(filename, mol, vertices_p, normals_p, indices_p, vertices_n, normals_n, indices_n)

def build_abo(filename, mol, vertices_p, normals_p, indices_p, vertices_n, normals_n, indices_n):
    outfile = ('%s.abo' % filename)
    f = open(outfile, 'wb')

    # write number of frames
    f.write(int(1).to_bytes(2, byteorder='little'))

    # write frame index
    f.write(int(1).to_bytes(2, byteorder='little'))

    # write description
    description = 'Orbital generated by PyQInt / PyTessel'
    f.write(len(description).to_bytes(2, byteorder='little'))
    f.write(bytearray(description, encoding='utf8'))

    # write atoms
    f.write(len(mol.charges).to_bytes(2, byteorder='little'))

    # loop over atoms
    for idx,pos in zip(mol.charges, mol.atoms):
        f.write(idx.to_bytes(1, byteorder='little'))
        f.write((np.array(pos[1], dtype=np.float32) * 0.529177).tobytes())

    # set colors
    color_pos = np.array([201, 2, 65, 0xF0]) / 255.0
    color_neg = np.array([2, 182, 201, 0xF0]) / 255.0

    # write number of models
    f.write(int(2).to_bytes(2, byteorder='little'))

    # write model index
    f.write(int(0).to_bytes(2, byteorder='little'))

    # write model color
    f.write(np.array(color_pos, dtype=np.float32).tobytes())

    # write number of vertices
    f.write(len(vertices_p).to_bytes(4, byteorder='little'))

    # loop over vertices and normals
    for v,n in zip(vertices_p, normals_p):
        f.write(np.array(v, dtype=np.float32).tobytes())
        f.write(np.array(n, dtype=np.float32).tobytes())

    # write number of faces2
    f.write((int(len(indices_p) / 3)).to_bytes(4, byteorder='little'))

    # write faces
    f.write(np.array(indices_p, dtype=np.uint32).tobytes())

    # write model index
    f.write(int(1).to_bytes(2, byteorder='little'))

    # write model color
    f.write(np.array(color_neg, dtype=np.float32).tobytes())

    # write number of vertices
    f.write(len(vertices_n).to_bytes(4, byteorder='little'))

    # loop over vertices and normals
    for v,n in zip(vertices_n, normals_n):
        f.write(np.array(v, dtype=np.float32).tobytes())
        f.write(np.array(n, dtype=np.float32).tobytes())

    # write indices
    f.write((int(len(indices_n) / 3)).to_bytes(4, byteorder='little'))

    # write faces
    f.write(np.array(indices_n, dtype=np.uint32).tobytes())

    f.close()



import unittest
from pyqint import PyQInt, Molecule
import numpy as np
import os

class TestDipole(unittest.TestCase):

    def test_cgf_dipole(self):
        """
        Test dipole integrals for contracted Gaussians

        Dijk = <cgf_i | d_k | cgf_j>
        """

        # construct integrator object
        integrator = PyQInt()

        # build water molecule
        mol = Molecule("H2O")
        mol.add_atom('O',  0.00000, -0.07579, 0.0000, unit='angstrom')
        mol.add_atom('H',  0.86681,  0.60144, 0.0000, unit='angstrom')
        mol.add_atom('H', -0.86681,  0.60144, 0.0000, unit='angstrom')
        cgfs, nuclei = mol.build_basis('sto3g')

        N = len(cgfs)
        D = np.zeros((N,N,3))
        for i in range(N):
            for j in range(i,N):
                for k in range(0,3):
                    D[i,j,k] = integrator.dipole(cgfs[i], cgfs[j], k)

        # test dipole integrals
        for i,cart in enumerate(['x','y','z']):
            exact = np.load(os.path.join(os.path.dirname(__file__),
                                         'results',
                                         'h2o_dipole_%s.npy' % cart))
            np.testing.assert_almost_equal(D[:,:,i], exact, decimal=4)

if __name__ == '__main__':
    unittest.main()
